import json
from jsonschema import validate
import sys 
import os



def load_schema(schema_name):
    schema_path = get_schema_path(schema_name)
    with open(schema_path, 'r') as schema_file:
        schema = json.load(schema_file)
    return schema


def get_schema_path(filename):
    """Get the path to a schema file within the metadata_templates package."""
    
    # Method 1: Try PyInstaller bundle first (onefolder creates _MEIPASS)
    if hasattr(sys, '_MEIPASS'):
        # PyInstaller onefolder extracts to _MEIPASS/
        possible_paths = [
            os.path.join(sys._MEIPASS, "pysoda", "schema", filename),
            os.path.join(sys._MEIPASS, filename)
        ]
        for path in possible_paths:
            if os.path.exists(path):
                return path
    
    # Method 2: Try to import the metadata_templates module (works if PyPI package is properly installed)
    try:
        from .. import schema
        schema_dir = os.path.dirname(schema.__file__)
        schema_path = os.path.join(schema_dir, filename)
        if os.path.exists(schema_path):
            return schema_path
    except (ImportError, ModuleNotFoundError, AttributeError):
        pass
    
    # Method 3: Search in the Flask app's directory structure
    current_file = os.path.abspath(__file__)
    current_dir = os.path.dirname(current_file)
    
    search_paths = [
        os.path.join(current_dir, '..', '..', 'schema', filename),
        os.path.join(current_dir, 'schema', filename),
    ]
    
    for path in search_paths:
        if os.path.exists(path):
            return path
        
    
    # Method 4: Use importlib_resources (Python 3.7+)
    try:
        from importlib import resources
        with resources.path('schema', filename) as schema_path:
            if schema_path.exists():
                return str(schema_path)
    except (ImportError, ModuleNotFoundError):
        # Fallback to other methods if importlib_resources is not available
        pass



        # Method 5: Try to find in Electron Resources folder
    try:
        # Find the Electron Resources folder
        current_path = current_dir
        resources_folder = None
        
        # Walk up the directory tree to find the Resources folder
        while current_path and current_path != os.path.dirname(current_path):
            # Check common Electron Resources locations
            possible_resources = [
                os.path.join(current_path, 'Resources'),  # macOS
                os.path.join(current_path, 'resources'),  # Windows/Linux
                os.path.join(current_path, 'Contents', 'Resources'),  # macOS app bundle
            ]
            
            for resource_path in possible_resources:
                if os.path.exists(resource_path):
                    resources_folder = resource_path
                    break
            
            if resources_folder:
                break
                
            current_path = os.path.dirname(current_path)
        
        # If we found the Resources folder, look for schema inside it
        if resources_folder:
            template_path = os.path.join(resources_folder, 'schema', filename)

            if os.path.exists(template_path):
                return template_path
                
    except Exception as e:
        pass
    
    raise ImportError(f"Could not locate or create schema file {filename}.")


# TODO: Make an enum of the schema names and add extensions to the schema names in the function.....or to the enum.
def validate_schema(schema, schema_name):
    """
    Validate submission metadata against the submission schema.

    Args:
        schema (dict): The python dictionary version of the schema or subschema to validate against the json schema.
        schema_name (str): The file name of the schema to validate against.

    Raises:
        ValidationError: If the metadata is invalid.
    """
    true_schema = load_schema(schema_name)
    validate(instance=schema, schema=true_schema)


def get_sds_headers(schema_name):
    """
    Get the headers for the SDS file.

    Args:
        soda (dict): The soda object containing the metadata.
        schema_name (str): The name of the schema to validate against.

    Returns:
        list: The headers for the SDS file.
    """

    true_schema = load_schema(schema_name)
    sds_headers = true_schema["items"][0]["properties"].keys()
    return sds_headers


from .submission import create_excel
from .dataset_description import create_excel
from .readme_changes import create_text_file
from .code_description import create_excel
from .manifest_package import create_high_level_manifest_files, get_auto_generated_manifest_files, load_metadata_to_dataframe, create_high_lvl_manifest_files_existing_ps_starting_point
from .manifest import create_excel, load_existing_manifest_file
from .resources import create_excel
from .performances import create_excel
from .submission import create_excel
from .sites import create_excel 
from .constants import (
    SDS_FILE_RESOURCES,
    SDS_FILE_PERFORMANCES,
    SDS_FILE_MANIFEST,
    SDS_FILE_SITES,
    SDS_FILE_CODE_DESCRIPTION,
    SDS_FILE_DATASET_DESCRIPTION,
    METADATA_UPLOAD_PS_PATH
)


from pennsieve2.pennsieve import Pennsieve
from .exceptions import PennsieveAgentError

def connect_pennsieve_client(account_name):
    """
        Connects to Pennsieve Python client to the Agent and returns the initialized Pennsieve object.
    """
    try:
        return Pennsieve(profile_name=account_name)
    except Exception as e:
        raise PennsieveAgentError(f"Could not connect to the Pennsieve agent: {e}")


import os
import requests
import pandas as pd
import itertools
from ...constants import PENNSIEVE_URL
from ...utils import create_request_headers, get_access_token, get_dataset_id, PennsieveActionNoPermission
from ..permissions import pennsieve_get_current_user_permissions
from ..metadata import load_metadata_to_dataframe

from .. import logger


METADATA_FILES_SPARC = [
        "submission.xlsx",
        "submission.csv",
        "submission.json",
        "dataset_description.xlsx",
        "dataset_description.csv",
        "dataset_description.json",
        "subjects.xlsx",
        "subjects.csv",
        "subjects.json",
        "samples.xlsx",
        "samples.csv",
        "samples.json",
        "README.txt",
        "CHANGES.txt",
        "code_description.xlsx",
        "inputs_metadata.xlsx",
        "outputs_metadata.xlsx",
        "manifest.xlsx",
        "manifest.csv"
    ]


def import_pennsieve_dataset(soda_json_structure, requested_sparc_only=True):
    global logger
    high_level_sparc_folders = [
        "code",
        "derivative",
        "docs",
        "primary",
        "protocol",
        "source",
    ]
    manifest_sparc = ["manifest.xlsx", "manifest.csv"]
    high_level_metadata_sparc = [
        "submission.xlsx",
        "submission.csv",
        "submission.json",
        "dataset_description.xlsx",
        "dataset_description.csv",
        "dataset_description.json",
        "subjects.xlsx",
        "subjects.csv",
        "subjects.json",
        "samples.xlsx",
        "samples.csv",
        "samples.json",
        "README.txt",
        "CHANGES.txt",
        "code_description.xlsx",
        "inputs_metadata.xlsx",
        "outputs_metadata.xlsx",
    ]
    double_extensions = [
        ".ome.tiff",
        ".ome.tif",
        ".ome.tf2,",
        ".ome.tf8",
        ".ome.btf",
        ".ome.xml",
        ".brukertiff.gz",
        ".mefd.gz",
        ".moberg.gz",
        ".nii.gz",
        ".mgh.gz",
        ".tar.gz",
        ".bcl.gz",
    ]

    global create_soda_json_completed
    global create_soda_json_total_items
    global create_soda_json_progress
    create_soda_json_progress = 0
    create_soda_json_total_items = 0
    create_soda_json_completed = 0

    # ["extensions"] doesn't seem to be returned by the Pennsieve API anymore
    def verify_file_name(file_name, extension):
        global logger
        if extension == "":
            return file_name

        double_ext = False
        for ext in double_extensions:
            if file_name.find(ext) != -1:
                double_ext = True
                break
                
        extension_from_name = ""

        if double_ext == False:
            extension_from_name = os.path.splitext(file_name)[1]
        else:
            extension_from_name = (
                os.path.splitext(os.path.splitext(file_name)[0])[1]
                + os.path.splitext(file_name)[1]
            )

        if extension_from_name == ("." + extension):
            return file_name
        else:
            return file_name + ("." + extension)


    def createFolderStructure(subfolder_json, manifest):
        """
            Function for creating the Pennsieve folder structure for a given dataset as an object stored locally.
            Arguments:
                subfolder_json: The json object containing the folder structure of the dataset
                pennsieve_client: The Pennsieve client object
                manifest: The manifest object for the dataset
        """
        # root level folder will pass subfolders into this function and will recursively check if there are subfolders while creating the json structure
        global logger
        global create_soda_json_progress
        
        collection_id = subfolder_json["path"]

        limit = 100
        offset = 0
        subfolder = []
        while True: 
            r = requests.get(f"{PENNSIEVE_URL}/packages/{collection_id}?limit={limit}&offset={offset}", headers=create_request_headers(get_access_token()))
            r.raise_for_status()
            page = r.json()["children"]
            subfolder.extend(page)
            if len(page) < limit:
                break
            offset += limit

        for items in subfolder:
            folder_item_name = items["content"]["name"]
            create_soda_json_progress += 1
            item_id = items["content"]["id"]
            # is a file name check if there are additional manifest information to attach to files
            if item_id[2:9] == "package":
                if (
                    folder_item_name[0:8] != "manifest"
                ):  # manifest files are not being included in json structure

                    # verify file name first (used for legacy Pennsieve datasets)
                    if("extension" not in subfolder):
                        folder_item_name = verify_file_name(folder_item_name, "")
                    else:
                        folder_item_name = verify_file_name(folder_item_name, subfolder["extension"])
                        
                    # verify timestamps
                    timestamp = items["content"]["createdAt"].replace('.', ',')

                    paths_list = [*subfolder_json["pspath"]]
                    subfolder_json["files"][folder_item_name] = {
                        "action": ["existing"],
                        "path": item_id,
                        "pspath": paths_list,
                        "timestamp": timestamp,
                        "location": "ps",
                        "additional-metadata": "",
                        "description": "",
                    }
                    
                    # creates path for folder_item_name (stored in temp_name)
                    if len(subfolder_json["files"][folder_item_name]["pspath"]) > 1:
                        temp_name = '/'.join(subfolder_json["files"][folder_item_name]["pspath"][1:]) + "/" + folder_item_name
                    else:
                        temp_name = folder_item_name
                    
                    if len(manifest.keys()) > 0:
                        # Dictionary that has the required manifest headers in lowercase and without spaces as keys
                        # and the correct manifest headers as values
                        defaultManifestHeadersNameMapped = {
                            "filename": "filename",
                            "timestamp": "timestamp",
                            "description": "description",
                            "filetype": "file type",
                            "entity": "entity",
                            "datamodality": "data modality",
                            "alsoindataset": "also in dataset",
                            "alsoindatasetpath": "also in dataset path",
                            "datadictionarypath": "data dictionary path",
                            "entityistransitive": "entity is transitive",
                            "additionalmetadata": "additional-metadata",
                        }

                        # Dictionary that will be used to store the correct manifest headers as keys
                        # and the values from the manifest as values
                        updated_manifest = {}

                        # Go through the imported manifest keys and change the keys to the correct name
                        # For example if the key is "File Name" change it to "filename"
                        for manifestKey in manifest.keys():
                            # Make the key lowercase
                            sterilizedKeyName = manifestKey.lower().replace(" ", "")
                            if sterilizedKeyName in defaultManifestHeadersNameMapped.keys():
                                # change the key to the correct name
                                # For example if the key name is "filetype" change it to "file type"
                                newManifestKeyName = defaultManifestHeadersNameMapped[sterilizedKeyName]
                                # Add the new key/value to the updated manifest
                                updated_manifest[newManifestKeyName] = manifest[manifestKey]
                            else:
                                # Keep the key/value the same and add it to the updated manifest
                                updated_manifest[manifestKey] = manifest[manifestKey]

                        if "filename" in updated_manifest.keys():
                            for manifestKey in updated_manifest.keys():
                                location_index = ""
                                # get the index of the file name in the manifest
                                if (temp_name in updated_manifest["filename"].values()):
                                    location_index = list(updated_manifest["filename"].values()).index(
                                        temp_name
                                    )
                                # This is for the case where the file name in the manifest has a slash at the beginning
                                # which is the case for files in the root folder
                                elif ("/" + temp_name in updated_manifest["filename"].values()):
                                    location_index = list(updated_manifest["filename"].values()).index(
                                        "/" + temp_name
                                    )
                                else:
                                    # break out of the for loop if the file name is not in the manifest
                                    break

                                # check if the key is in the required manifest headers, if it is, update the folder_item_name value
                                # corresponding to the key
                                if manifestKey in defaultManifestHeadersNameMapped.values():
                                    if updated_manifest[manifestKey][location_index] != "":
                                        if folder_item_name[0:1] == "/":
                                            subfolder_json["files"][folder_item_name[:1]][manifestKey] = updated_manifest[manifestKey][location_index]
                                        else:
                                            subfolder_json["files"][folder_item_name][manifestKey] = updated_manifest[manifestKey][location_index]
                                # if the key is not in the required manifest headers, add it to the extra columns folder_item_name value
                                else :
                                    # if the extra columns key does not exist, create it
                                    if "extra_columns" not in subfolder_json["files"][folder_item_name]:
                                        subfolder_json["files"][folder_item_name]["extra_columns"] = {}
                                    
                                    if updated_manifest[manifestKey][location_index] != "":
                                        subfolder_json["files"][folder_item_name]["extra_columns"][manifestKey] = updated_manifest[manifestKey][location_index]
                                    else:
                                        subfolder_json["files"][folder_item_name]["extra_columns"][manifestKey] = ""
                        else:
                            # filename not in updated manifest so recreate standard headers if they don't exist
                            # loop through the updated manifest keys and if header matches standard header add content else recreate
                            if len(updated_manifest.keys()) > 0:
                                location_index = ""
                                for manifestKey in updated_manifest.keys():
                                    if temp_name in updated_manifest[manifestKey].values():
                                        # file_names found
                                        location_index = list(updated_manifest[manifestKey].values()).index(
                                        temp_name
                                        )
                                    if ("/" + temp_name in updated_manifest[manifestKey].values()):
                                        location_index = list(updated_manifest[manifestKey].values()).index(
                                        "/" + temp_name
                                        )
                                    if location_index != "":
                                        if manifestKey in defaultManifestHeadersNameMapped.values():
                                            if folder_item_name[0:1] == "/":
                                                subfolder_json["files"][folder_item_name[1:]][manifestKey] = updated_manifest[manifestKey][location_index]
                                            else:
                                                subfolder_json["files"][folder_item_name][manifestKey] = updated_manifest[manifestKey][location_index]
                                        else:
                                            if "extra_columns" not in subfolder_json["files"][folder_item_name]:
                                                subfolder_json["files"][folder_item_name]["extra_columns"] = {}
                                            subfolder_json["files"][folder_item_name]["extra_columns"][manifestKey] = updated_manifest[manifestKey][location_index]

            else:  # another subfolder found
                paths_list = [*subfolder_json["pspath"], folder_item_name]
                subfolder_json["folders"][folder_item_name] = {
                    "action": ["existing"],
                    "path": item_id,
                    "pspath": paths_list,
                    "files": {},
                    "folders": {},
                    "location": "ps",
                }

        if len(subfolder_json["folders"].keys()) != 0:  # there are subfolders
            for folder in subfolder_json["folders"].keys():
                subfolder = subfolder_json["folders"][folder]
                createFolderStructure(subfolder, manifest)


    # check that the Pennsieve dataset is valid
    try:
        bf_dataset_name = soda_json_structure["ps-dataset-selected"]["dataset-name"]
    except Exception as e:
        raise e

    selected_dataset_id = get_dataset_id(bf_dataset_name)

    # check that the user has permission to edit this dataset
    role = pennsieve_get_current_user_permissions(selected_dataset_id, get_access_token())["role"]
    if role not in ["owner", "manager", "editor"]:
        curatestatus = "Done"
        raise PennsieveActionNoPermission("You do not have permissions to edit upload this Pennsieve dataset.")


    # surface layer of dataset is pulled. then go through through the children to get information on subfolders
    manifest_dict = {}
    manifest_error_message = []
    soda_json_structure["dataset-structure"] = {
        "files": {},
        "folders": {},
    }

    # root of dataset is pulled here (high level folders/files are gathered here)
    # root_folder is the files and folders within root
    r = requests.get(f"{PENNSIEVE_URL}/datasets/{selected_dataset_id}", headers=create_request_headers(get_access_token()))
    r.raise_for_status()
    root_folder = r.json()["children"]

    # Get the amount of files/folders in the dataset
    r = requests.get(f"{PENNSIEVE_URL}/datasets/{selected_dataset_id}/packageTypeCounts", headers=create_request_headers(get_access_token()))
    r.raise_for_status()
    packages_list = r.json()


    # root's children files
    for count in packages_list.values():
        create_soda_json_total_items += int(count)

    # set manifest dictionry to empty dictionary; used to store the manifest information while we import dataset
    manifest_dict = {}
    

    # Gather metadata files first
    for items in root_folder:
        item_id = items["content"]["id"]
        item_name = items["content"]["name"]

        # Import manifest at the root of the dataset
        if item_name in manifest_sparc:
            # Item is a manifest file
            df = ""
            try:
                if item_name.lower() == "manifest.xlsx":
                    df = load_metadata_to_dataframe(item_id, "excel", get_access_token())
                    df = df.fillna("")
                else:
                    df = load_metadata_to_dataframe(item_id, "csv", get_access_token())
                    df = df.fillna("")
                manifest_dict = df.to_dict()
            except Exception as e:
                manifest_error_message.append(item_name)

        # Item is a metadata file
        if item_name in high_level_metadata_sparc:
            create_soda_json_progress += 1
            if "dataset_metadata" not in soda_json_structure.keys():
                soda_json_structure["dataset_metadata"] = {}
            soda_json_structure["dataset_metadata"][item_name] = {
                "location": "ps",
                "action": ["existing"],
                "path": item_id,
            }

    # Process the folder structure
    for items in root_folder:
        item_id = items["content"]["id"]
        item_name = items["content"]["name"]

        # If package type is Collection, then it is a folder
        if items["content"]["packageType"] == "Collection" and item_name in high_level_sparc_folders:
            create_soda_json_progress += 1
            soda_json_structure["dataset-structure"]["folders"][item_name] = {
                "location": "ps",
                "path": item_id,
                "action": ["existing"],
                "files": {},
                "folders": {},
                "pspath": [item_name],
            }

            # Check the content of the folder to see if a manifest file exists
            r = requests.get(f"{PENNSIEVE_URL}/packages/{item_id}", headers=create_request_headers(get_access_token()))
            r.raise_for_status()
            folder_content = r.json()["children"]

            if len(folder_content) > 0:
                high_lvl_folder_dict = soda_json_structure["dataset-structure"]["folders"][item_name]

                createFolderStructure(
                    high_lvl_folder_dict, manifest_dict
                )  # Passing item's JSON and the collection ID

    success_message = (
        "Data files under a valid high-level SPARC folders have been imported"
    )
    create_soda_json_completed = 1

    logger.info(f"Time to import {soda_json_structure['ps-dataset-selected']['dataset-name']} ")
    return {
        "soda_object": soda_json_structure,
        "success_message": success_message,
        "manifest_error_message": manifest_error_message,
        "import_progress": create_soda_json_progress,
        "import_total_items": create_soda_json_total_items,
    }


create_soda_json_progress = 0
create_soda_json_total_items = 0
create_soda_json_completed = 0


def create_soda_json_object_backend(
    soda_json_structure, root_folder_path, irregularFolders, replaced
):
    """
    This function is meant for importing local datasets into SODA.
    It creates a json object with the structure of the dataset.
    """
    global create_soda_json_progress  # amount of items counted during recursion
    global create_soda_json_total_items  # counts the total items in folder
    global create_soda_json_completed  # completed progress is either 0 or 1
    global METADATA_FILES_SPARC

    high_level_sparc_folders = [
        "code",
        "derivative",
        "docs",
        "primary",
        "protocol",
        "source",
    ]

    dataset_folder = soda_json_structure["dataset-structure"] = {"folders": {}}

    def recursive_structure_create(dataset_structure, folder_path, root_manifest):
        global create_soda_json_progress
        # going within high level folders
        # add manifest details if manifest exists
        manifest_object = {
            "filename": "",
            "timestamp": "",
            "description": "",
            "file type": "",
            "entity": "",
            "data modality": "",
            "also in dataset": "",
            "also in dataset path": "",
            "data dictionary path": "",
            "entity is transitive": "",
            "additional-metadata": "",
        }

        lastSlash = folder_path.rfind("/") + 1
        folder_name = folder_path[lastSlash:]

        if folder_name in replaced.keys():
            folder_name = replaced[folder_name]

        # Check if folder is in irregular folders
        if folder_path in irregularFolders:
            index_check = irregularFolders.index(folder_path)
            modified_name = replaced[os.path.basename(folder_path)]
            folder_path = irregularFolders[index_check]


        entries = os.listdir(folder_path)
        for entry in entries:
            item_path = os.path.normpath(os.path.join(folder_path, entry))
            if os.path.isfile(item_path):
                # Check manifest to add metadata
                if entry[0:1] != "." and entry[0:8] != "manifest":
                    create_soda_json_progress += 1
                    # Use the root manifest to find metadata for the file
                    for row in root_manifest:
                        extra_columns = False
                        if len(row) > 11:
                            extra_columns = True
                            extra_columns_dict = dict(itertools.islice(row.items(), 5, len(row)))

                        if row["filename"] == entry:
                            # Add description metadata
                            manifest_object["description"] = row.get("description", "")
                            # Add additional metadata
                            manifest_object["additional-metadata"] = row.get("Additional Metadata", "")
                            if extra_columns:
                                manifest_object["extra_columns"] = extra_columns_dict

                    # Create JSON structure for the file
                    if "extra_columns" in manifest_object:
                        dataset_structure["files"][entry] = {
                            "path": item_path,
                            "location": "local",
                            "action": ["existing"],
                            "description": manifest_object["description"],
                            "additional-metadata": manifest_object["additional-metadata"],
                            "extra_columns": manifest_object["extra_columns"],
                        }
                    else:
                        dataset_structure["files"][entry] = {
                            "path": item_path,
                            "location": "local",
                            "action": ["existing"],
                            "description": manifest_object["description"],
                            "additional-metadata": manifest_object["additional-metadata"],
                        }
            elif os.path.isdir(item_path) is True:
                create_soda_json_progress += 1
                if item_path in irregularFolders:
                    index_check = irregularFolders.index(item_path)
                    modified_name = replaced[os.path.basename(item_path)]

                    dataset_structure["folders"][modified_name] = {
                        "folders": {},
                        "files": {},
                        "path": item_path,
                        "location": "local",
                        "action": ["existing"],
                        "original-name": entry,
                    }
                    for folder in dataset_structure["folders"][modified_name][
                        "folders"
                    ]:
                        updated_path = dataset_structure["folders"][modified_name][
                            folder
                        ]["path"]
                        recursive_structure_create(
                            dataset_structure["folders"][modified_name][folder],
                            updated_path,
                            root_manifest
                        )
                else:
                    dataset_structure["folders"][entry] = {
                        "folders": {},
                        "files": {},
                        "path": item_path,
                        "location": "local",
                        "action": ["existing"],
                    }

        for folder in dataset_structure["folders"]:
            updated_path = dataset_structure["folders"][folder]["path"]
            recursive_structure_create(
                dataset_structure["folders"][folder], updated_path, root_manifest
            )

    # BEGIN

    # Check for a single manifest file at the root of the dataset
    root_manifest_csv = os.path.join(root_folder_path, "manifest.csv")
    root_manifest_xlsx = os.path.join(root_folder_path, "manifest.xlsx")

    soda_json_structure["starting-point"]["manifest"] = {}

    if os.path.exists(root_manifest_csv):
        csv_data = pd.read_csv(root_manifest_csv)
        csv_data.fillna("", inplace=True)
        json_format = csv_data.to_dict(orient="records")
        soda_json_structure["starting-point"]["manifest"] = json_format
        soda_json_structure["starting-point"]["path"] = root_manifest_csv
    elif os.path.exists(root_manifest_xlsx):
        excel_data = pd.read_excel(root_manifest_xlsx, sheet_name="Sheet1")
        excel_data.fillna("", inplace=True)
        json_format = excel_data.to_dict(orient="records")
        soda_json_structure["starting-point"]["manifest"] = json_format
        soda_json_structure["starting-point"]["path"] = root_manifest_xlsx



    # count the amount of items in folder
    create_soda_json_total_items = 0
    for root, dirs, filenames in os.walk(root_folder_path):
        # walk through all folders and it's subfolders
        for Dir in dirs:
            # does not take hidden folders or manifest folders
            if Dir[0:1] != "." and Dir[0:8] != "manifest":
                create_soda_json_total_items += 1
        for fileName in filenames:
            if root == root_folder_path and fileName in METADATA_FILES_SPARC:
                # goes through all files and does not count hidden files
                create_soda_json_total_items += 1
            else:
                if fileName[0:1] != ".":
                    create_soda_json_total_items += 1

    # reading high level folders
    create_soda_json_completed = 0
    create_soda_json_progress = 0
    entries = os.listdir(root_folder_path)


    for entry in entries:
        # begin going through high level folders
        item_path = os.path.normpath(os.path.join(root_folder_path, entry))
        # high level folder paths
        if os.path.isfile(item_path) is True:
            if entry[0:1] != "." and entry in METADATA_FILES_SPARC:
                # is not a hidden folder
                create_soda_json_progress += 1
                soda_json_structure["dataset_metadata"][entry] = {
                    "path": item_path,
                    "location": "local",
                    "action": ["existing"],
                }
            # do file work here
        elif os.path.isdir(item_path) is True:
            create_soda_json_progress += 1
            # add item to soda
            if item_path in irregularFolders:
                index_check = irregularFolders.index(item_path)
                modified_name = replaced[index_check]
                folder_name = modified_name
                dataset_folder["folders"][folder_name] = {
                    "folders": {},
                    "files": {},
                    "path": item_path,
                    "location": "local",
                    "action": ["existing"],
                    "original-basename": item_path[(item_path.rfind("/") + 1) :],
                }
            else:
                if entry in high_level_sparc_folders:
                    dataset_folder["folders"][entry] = {
                        "folders": {},
                        "files": {},
                        "path": item_path,
                        "location": "local",
                        "action": ["existing"],
                    }
            soda_json_structure["starting-point"][entry] = {"path": ""}

    for folder in dataset_folder["folders"]:
        # go through high level folders again
        high_lvl_path = root_folder_path + "/" + folder
        recursive_structure_create(dataset_folder["folders"][folder], high_lvl_path, soda_json_structure["starting-point"]["manifest"])

    create_soda_json_completed = 1
    return soda_json_structure


def monitor_local_json_progress():
    """
    Function for monitoring progress of json_object_creation
    Used for progress bar
    """
    global create_soda_json_completed
    global create_soda_json_total_items
    global create_soda_json_progress
    progress_percentage = (
        create_soda_json_progress / create_soda_json_total_items
    ) * 100

    return {
        "create_soda_json_progress": create_soda_json_progress,
        "create_soda_json_total_items": create_soda_json_total_items,
        "progress_percentage": progress_percentage,
        "create_soda_json_completed": create_soda_json_completed
    }



from .upload import (
    create_folder_level_manifest,
    check_empty_files_folders,
    main_curate_function,
    main_curate_function_progress,
    generate_manifest_file_locally,
    generate_manifest_file_data,
    check_json_size,
    clean_json_structure,
    check_server_access_to_files,
)


from soda_object import soda as soda_object

class Soda:
    def __init__(self, standard=""):
        self.soda_object = soda_object
        self.soda_object["standard"] = standard

    def set_standard(self, standard):
        self.soda_object["standard"] = standard

    def get_soda_object(self):
        return self.soda_object

    # Add more methods to alter the soda_object as needed
    def update_metadata(self, key, value):
        if key in self.soda_object["dataset-metadata"]:
            self.soda_object["dataset-metadata"][key] = value
        else:
            raise KeyError(f"{key} not found in dataset-metadata")

    def add_performance(self, performance):
        self.soda_object["performances"].append(performance)

    def add_subject_metadata(self, metadata):
        self.soda_object["subject-metadata"].append(metadata)

    def set_dataset_name(self, name):
        self.soda_object["dataset-metadata"]["name"] = name

    def get_dataset_name(self):
        return self.soda_object["dataset-metadata"]["name"]

    def get_submission(self):
        return self.soda_object["dataset-metadata"]["submission-metadata"]


"""
Purpose: Functions for dealing with the pennsieve config file located at ~/.pennsieve/config.ini 
"""
from ..constants import PENNSIEVE_URL


def add_api_host_to_config(configparser, target_section_name, configpath):
    """
    Args:
        configparser: configparser object
        target_section_name: the section name to add the api_host to. Should be an account section. 
        configpath: the path to the config file ( ~/.pennsieve/config.ini )
    Action:
        Adds api_host to the section of the configparser object if it does not exist in the given section.
        If given a section name that is 'agent' it will do nothing.
    """

    # do not add the api_host key to the agent section of the config
    if target_section_name == 'agent':
        return 

    if not configparser.has_option(target_section_name, "api_host"):
        configparser.set(target_section_name, "api_host", PENNSIEVE_URL)

    with open(configpath, "w+") as configfile:
        configparser.write(configfile)



def format_agent_profile_name(profile_name):
    """
    Args:
        profile_name: the profile name to format
    Returns:
        The formatted profile name
    Notes:
        We replace the '.' with '_' because the config parser on the Node side does not like '.' in the profile name.
    """
    return profile_name.lower().replace('.', '_').strip()


def lowercase_account_names(config, account_name, configpath):
    """
    Args:
        config: configparser object
        account_name: the account name to convert to lowercase
    Action:
        Converts the account name and global default_profile value to lowercase and updates the config file.
    """
    formatted_account_name = format_agent_profile_name(account_name)
    # if the section exists lowercased do nothing 
    if config.has_section(formatted_account_name):
        return

    # add the section back with the lowercase account name
    config.add_section(formatted_account_name) 
    config.set(formatted_account_name, "api_token", config.get(account_name, "api_token"))
    config.set(formatted_account_name, "api_secret", config.get(account_name, "api_secret"))

    # set the global default_profile option to lowercase
    config.set("global", "default_profile", formatted_account_name)

    # remove the unformatted account name
    config.remove_section(account_name)

    # finalize the changes
    with open(configpath, "w+") as configfile:
        config.write(configfile)


import requests
from ...constants import PENNSIEVE_URL
from ...utils import get_dataset_id, get_access_token, create_request_headers, connect_pennsieve_client, PennsieveActionNoPermission, GenericUploadError
from ...core import has_edit_permissions
from functools import partial
import time 
import os
from .. import logger
import sys


def get_template_path(filename):
    """Get the path to a template file within the metadata_templates package."""
    global logger
    
    # Method 1: Try PyInstaller bundle first (onefolder creates _MEIPASS)
    if hasattr(sys, '_MEIPASS'):
        # PyInstaller onefolder extracts to _MEIPASS/
        possible_paths = [
            os.path.join(sys._MEIPASS, "pysoda", "core", "metadata_templates", filename),
            os.path.join(sys._MEIPASS, "metadata_templates", filename),
            os.path.join(sys._MEIPASS, filename)
        ]
        for path in possible_paths:
            if os.path.exists(path):
                logger.info(f"Template found in PyInstaller bundle: {path}")
                return path
    
    # Method 2: Try to import the metadata_templates module (works if PyPI package is properly installed)
    try:
        from .. import metadata_templates
        templates_dir = os.path.dirname(metadata_templates.__file__)
        template_path = os.path.join(templates_dir, filename)
        if os.path.exists(template_path):
            logger.info(f"Template found in metadata_templates module: {template_path}")
            return template_path
    except (ImportError, ModuleNotFoundError, AttributeError):
        pass
    
    # Method 3: Search in the Flask app's directory structure
    current_file = os.path.abspath(__file__)
    current_dir = os.path.dirname(current_file)
    
    # Walk up the directory tree to find the templates
    search_paths = [
        os.path.join(current_dir, '..', 'metadata_templates', filename),
        os.path.join(current_dir, 'metadata_templates', filename),
    ]
    
    # Also check if we're in a site-packages structure
    site_packages_paths = []
    path_parts = current_file.split(os.sep)
    for i, part in enumerate(path_parts):
        if part == 'site-packages':
            site_packages_root = os.sep.join(path_parts[:i+1])
            site_packages_paths.extend([
                os.path.join(site_packages_root, 'pysoda', 'core', 'metadata_templates', filename),
                os.path.join(site_packages_root, 'pysoda_fairdataihub_tools', 'pysoda', 'core', 'metadata_templates', filename)
            ])
    
    all_paths = search_paths + site_packages_paths
    
    for path in all_paths:
        if os.path.exists(path):
            logger.info(f"Template found in directory structure: {path}")
            return path
    
    # Method 4: Try to find in Electron app resources (if not using PyInstaller)
    try:
        # Look for Electron app structure
        current_path = current_dir
        while current_path and current_path != os.path.dirname(current_path):
            electron_paths = [
                os.path.join(current_path, 'resources', 'app', 'node_modules', 'pysoda', 'core', 'metadata_templates', filename),
                os.path.join(current_path, 'resources', 'pysoda', 'core', 'metadata_templates', filename),
                os.path.join(current_path, 'app', 'pysoda', 'core', 'metadata_templates', filename)
            ]
            for path in electron_paths:
                if os.path.exists(path):
                    logger.info(f"Template found in Electron app resources: {path}")
                    return path
            current_path = os.path.dirname(current_path)
    except Exception:
        pass


    # Method 5: Try to find in Electron Resources folder
    try:
        # Find the Electron Resources folder
        current_path = current_dir
        resources_folder = None
        
        # Walk up the directory tree to find the Resources folder
        while current_path and current_path != os.path.dirname(current_path):
            # Check common Electron Resources locations
            possible_resources = [
                os.path.join(current_path, 'Resources'),  # macOS
                os.path.join(current_path, 'resources'),  # Windows/Linux
                os.path.join(current_path, 'Contents', 'Resources'),  # macOS app bundle
            ]
            
            for resource_path in possible_resources:
                if os.path.exists(resource_path):
                    resources_folder = resource_path
                    break
            
            if resources_folder:
                break
                
            current_path = os.path.dirname(current_path)
        
        # If we found the Resources folder, look for metadata_templates inside it
        if resources_folder:
            template_path = os.path.join(resources_folder, 'metadata_templates', filename)
            logger.info(f"Searching for template file in Electron Resources: {template_path}")

            if os.path.exists(template_path):
                logger.info(f"Template found in Electron Resources: {template_path}")
                return template_path
                
    except Exception as e:
        logger.warning(f"Failed to search Electron Resources: {e}")
        pass

    # Method 6: Use importlib_resources as fallback (Python 3.7+)
    try:
        from importlib import resources
        with resources.path('metadata_templates', filename) as template_path:
            logger.info(f"Using template path: {template_path}")

            if template_path.exists():
                logger.info(f"Template found using importlib_resources: {template_path}")
                return str(template_path)
    except (ImportError, ModuleNotFoundError, AttributeError):
        # Fallback to other methods if importlib_resources is not available
        pass
    

        
    except Exception as e:
        logger.error(f"Failed to create fallback template: {e}")
        raise ImportError(f"Could not locate or create template file {filename}. Error: {e}")



# helper function to process custom fields (users add and name them) for subjects and samples files
def getMetadataCustomFields(matrix):
    return [column for column in matrix if any(column[1:])]


# transpose a matrix (array of arrays)
# The transpose of a matrix is found by interchanging its rows into columns or columns into rows.
# REFERENCE: https://byjus.com/maths/transpose-of-a-matrix/
def transposeMatrix(matrix):
    return [[matrix[j][i] for j in range(len(matrix))] for i in range(len(matrix[0]))]

# needed to sort subjects and samples table data to match the UI fields
def sortedSubjectsTableData(matrix, fields):
    sortedMatrix = []
    for field in fields:
        for column in matrix:
            if column[0].lower() == field:
                sortedMatrix.append(column)
                break

    customHeaderMatrix = [
        column for column in matrix if column[0].lower() not in fields
    ]

    return (
        np.concatenate((sortedMatrix, customHeaderMatrix)).tolist()
        if customHeaderMatrix
        else sortedMatrix
    )



def upload_metadata_file(file_name, soda, path_to_file, delete_after_upload=True):
    global logger

    if "ps-account-selected" in soda:
        ps_account = soda["ps-account-selected"]["account-name"]
    
    if "ps-dataset-selected" in soda:
        ps_dataset = soda["ps-dataset-selected"]["dataset-name"]
    
    # check that the Pennsieve dataset is valid
    selected_dataset_id = get_dataset_id(ps_dataset)

    # check that the user has permissions for uploading and modifying the dataset
    if not has_edit_permissions(get_access_token(), selected_dataset_id):
        raise PennsieveActionNoPermission("edit" + selected_dataset_id)
    headers = create_request_headers(get_access_token())
    # handle duplicates on Pennsieve: first, obtain the existing file ID
    r = requests.get(f"{PENNSIEVE_URL}/datasets/{selected_dataset_id}", headers=headers)
    r.raise_for_status()
    ds_items = r.json()
    # go through the content in the dataset and find the file ID of the file to be uploaded
    for item in ds_items["children"]:
        if item["content"]["name"] == file_name:
            item_id = item["content"]["id"]
            jsonfile = {
                "things": [item_id]
            }
            # then, delete it using Pennsieve method delete(id)\vf = Pennsieve()
            r = requests.post(f"{PENNSIEVE_URL}/data/delete",json=jsonfile, headers=headers)
            r.raise_for_status()
    try:
        ps = connect_pennsieve_client(ps_account)
        # create a new manifest for the metadata file
        ps.use_dataset(selected_dataset_id)
        manifest = ps.manifest.create(path_to_file)
        m_id = manifest.manifest_id
    except Exception as e:
        logger.error(e)
        error_message = "Could not create manifest file for this dataset"
        raise GenericUploadError(error_message)
    
    # upload the manifest file
    try: 
        ps.manifest.upload(m_id)
        # create a subscriber function with ps attached so it can be used to unusbscribe
        subscriber_metadata_ps_client = partial(subscriber_metadata, ps)
        # subscribe for the upload to finish
        ps.subscribe(10, False, subscriber_metadata_ps_client)
    except Exception as e:
        logger.error("Error uploading dataset files")
        logger.error(e)
        raise Exception("The Pennsieve Agent has encountered an issue while uploading. Please retry the upload. If this issue persists please follow this <a target='_blank' rel='noopener noreferrer' href='https://docs.sodaforsparc.io/docs/how-to/how-to-reinstall-the-pennsieve-agent'> guide</a> on performing a full reinstallation of the Pennsieve Agent to fix the problem.")


    # before we can remove files we need to wait for all of the Agent's threads/subprocesses to finish
    # elsewise we get an error that the file is in use and therefore cannot be deleted
    time.sleep(5)

    # delete the local file that was created for the purpose of uploading to Pennsieve
    if delete_after_upload:
        os.remove(path_to_file)



def subscriber_metadata(ps, events_dict):
    global logger
    if events_dict["type"] == 1:
        fileid = events_dict["upload_status"].file_id
        total_bytes_to_upload = events_dict["upload_status"].total
        current_bytes_uploaded = events_dict["upload_status"].current
        if current_bytes_uploaded == total_bytes_to_upload and fileid != "":
            logger.info("File upload complete")
            ps.unsubscribe(10)


from pennsieve2 import Pennsieve
import re
import math



class UploadManifestSession:

    # properties 
    df_mid = None
    ps = None

    # upload values 
    main_total_generate_dataset_size = None
    total_files_to_upload = None 
    elapsed_time = None

    # rename values
    renaming_files_flow = False
    rename_total_files = None
    list_of_files_to_rename = None

    def __init__(self):
        self.df_mid = None

    def set_df_mid(self, id):
        self.df_mid = id
    
    def get_df_mid(self):
        return self.df_mid
    
    def set_elapsed_time(self, time):
        self.elapsed_time = time

    def get_elapsed_time(self):
        return self.elapsed_time
    
    def set_main_total_generate_dataset_size(self, size):
        self.main_total_generate_dataset_size = size

    def get_main_total_generate_dataset_size(self):
        return self.main_total_generate_dataset_size
    
    def set_total_files_to_upload(self, count):
        self.total_files_to_upload = count

    def get_total_files_to_upload(self):
        return self.total_files_to_upload

    def set_rename_total_files(self, count):
        self.rename_total_files = count

    def get_rename_total_files(self):
        return self.rename_total_files
    
    def set_list_of_files_to_rename(self, list):
        self.list_of_files_to_rename = list

    def get_list_of_files_to_rename(self):
        return self.list_of_files_to_rename
    
    def set_renaming_files_flow(self, value):
        self.renaming_files_flow = value

    def get_renaming_files_flow(self):
        return self.renaming_files_flow

    def df_mid_has_progress(self):
        if self.ps is None:
            self.ps = Pennsieve()
        try: 
            self.ps.manifest.sync(self.df_mid)
        except Exception as e:
            return False

        try: 
            mfs = self.ps.list_manifests()
        except Exception as e:
            # there are no manifests created yet
            return False
        return any(mf.id == self.df_mid and mf.status == "Initiated" for mf in mfs)
    
    def get_remaining_file_count(self, mid, total_files):
        if self.ps is None:
            self.ps = Pennsieve()
        total_pages = math.ceil(total_files / 1000)
        remaining_files = 0
        offset = 0
        for i in range(total_pages):
            if i >= 1:
                offset += 1000
            file_page = self.ps.manifest.list_files(mid, offset , 1000)
            # if there is no node_id then an upload hasn't started yet - all files are remaining 
            # regular expression that searches and counts for every string that has "status: LOCAL" or "status: REGISTERED" or "status: FAILED" in the string
            remaining_files +=  len(re.findall(r'status: REGISTERED|status: LOCAL|status: FAILED' , str(file_page)))
        return remaining_files
    
    def create_obj_from_string(self,s):
        # Split into individual objects
        objects = re.findall(r'file {([^}]*?)}', s, re.DOTALL)

        # Parse each object
        parsed_objects = []
        for obj in objects:
            # Split into lines and remove empty lines
            lines = [line.strip() for line in obj.split('\n') if line.strip()]
            # Split each line into key and value and create a dictionary
            parsed_object = {line.split(': ')[0]: line.split(': ')[1] for line in lines}
            parsed_objects.append(parsed_object)

        return parsed_objects

    def calculate_completed_upload_size(self, mid, bytes_per_file_dict, total_files):
        if self.ps is None:
            self.ps = Pennsieve()
        total_pages = math.ceil(total_files / 1000)
        offset = 0
        total_bytes_uploaded = 0
        for i in range(total_pages):
            if i >= 1:
                offset += 1000
            file_string = self.ps.manifest.list_files(mid, offset , 1000)
            parsed_objects = self.create_obj_from_string(str(file_string))
            for obj in parsed_objects:
                if 'status' not in obj:
                    total_bytes_uploaded += 0
                elif obj['status'] in [
                    'UPLOADED',
                    'IMPORTED',
                    'FINALIZED',
                    'VERIFIED',
                ]:
                    file_path = obj['source_path']
                    # remove the first and last characer of file_path - these are quotation marks
                    file_path = file_path[1:-1]
                    total_bytes_uploaded += int(bytes_per_file_dict.get(file_path, 0))

        return total_bytes_uploaded

    




# ums = UploadManifestSession()
# ums.df_mid_has_progress()


from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH, SDS_FILE_DATASET_DESCRIPTION, SCHEMA_NAME_DATASET_DESCRIPTION
from os.path import join, getsize
from openpyxl import load_workbook
import shutil
from .excel_utils import rename_headers, excel_columns
import itertools
from openpyxl.styles import PatternFill
from ...utils import validate_schema
from .helpers import upload_metadata_file, get_template_path



def create_excel(
    soda,
    upload_boolean,
    local_destination,
):
    source = get_template_path(SDS_FILE_DATASET_DESCRIPTION)
    destination = join(METADATA_UPLOAD_PS_PATH, SDS_FILE_DATASET_DESCRIPTION) if upload_boolean else local_destination
    shutil.copyfile(source, destination)

    validate_schema(soda["dataset_metadata"]["dataset_description"], SCHEMA_NAME_DATASET_DESCRIPTION)

    # write to excel file
    wb = load_workbook(destination)
    ws1 = wb["Sheet1"]

    ws1["D22"] = ""
    ws1["E22"] = ""
    ws1["D24"] = ""
    ws1["E24"] = ""
    ws1["D25"] = ""
    ws1["E25"] = ""

    # Populate the Metadata version (Required)
    ws1["D2"] = soda["dataset_metadata"]["dataset_description"]["metadata_version"]

    # Populate the Dataset Type (default to empty string if not present)
    ws1["D3"] = (
    soda.get("dataset_metadata", {})
        .get("dataset_description", {})
        .get("dataset_type", "")
    )

    populate_standards_info(ws1, soda)

    keyword_array = populate_basic_info(ws1, soda)

    populate_study_info(ws1, soda)
    populate_contributor_info(ws1, soda)
    populate_related_resource_information(ws1, soda)
    populate_funding_info(ws1, soda)
    populate_participant_information(ws1, soda)
    data_dictionary_information(ws1, soda)

    wb.save(destination)

    size = getsize(destination)

    ## if generating directly on Pennsieve, then call upload function and then delete the destination path
    if upload_boolean:
        upload_metadata_file(
            "dataset_description.xlsx", soda, destination, True
        )

    return {"size": size}




def populate_study_info(workbook, soda):
    study_info = soda["dataset_metadata"]["dataset_description"]["study_information"]
    workbook["D20"] = study_info.get("study_purpose", "")
    workbook["D21"] = study_info.get("study_data_collection", "")
    workbook["D22"] = study_info.get("study_primary_conclusion", "")

    # Arrays
    organ_system = study_info.get("study_organ_system", [])
    approach = study_info.get("study_approach", [])
    technique = study_info.get("study_technique", [])

    for i, column in zip(range(len(organ_system)), excel_columns(start_index=3)):
        workbook[column + "23"] = organ_system[i]
    for i, column in zip(range(len(approach)), excel_columns(start_index=3)):
        workbook[column + "24"] = approach[i]
    for i, column in zip(range(len(technique)), excel_columns(start_index=3)):
        workbook[column + "25"] = technique[i]

    workbook["D26"] = study_info.get("study_collection_title", "")

    # Return the max length of the arrays, or 1 if all are empty
    return max(1, len(organ_system), len(approach), len(technique))

def populate_standards_info(workbook, soda):
    standards_info = soda["dataset_metadata"]["dataset_description"]["standards_information"]
    workbook["D5"] = standards_info["data_standard"]
    workbook["D6"] = standards_info["data_standard_version"]


def populate_basic_info(workbook, soda):
    basic_info = soda["dataset_metadata"]["dataset_description"]["basic_information"]
    workbook["D8"] = basic_info.get("title", "")
    workbook["D9"] = basic_info.get("subtitle", "")
    workbook["D10"] = basic_info.get("description", "")

    # Write keywords array across columns in row 11 (D11, E11, F11, ...)
    keywords = basic_info.get("keywords", [])
    for col, keyword in zip(excel_columns(start_index=3), keywords):
        workbook[f"{col}11"] = keyword

    workbook["D12"] = basic_info.get("funding", "")
    workbook["D13"] = basic_info.get("acknowledgments", "")
    workbook["D14"] = basic_info.get("license", "")

    # Return the length of the keywords array, or 1 if empty
    return max(1, len(keywords))


def populate_funding_info(workbook, soda):
    funding_info = soda["dataset_metadata"]["dataset_description"]["funding_information"]
    workbook["D16"] = funding_info["funding_consortium"]
    workbook["D17"] = funding_info["funding_agency"]
    workbook["D18"] = funding_info["award_number"]



def populate_contributor_info(workbook, soda):
    contributor_info = soda["dataset_metadata"]["dataset_description"].get("contributor_information", [])
    for contributor, column in zip(contributor_info, excel_columns(start_index=3)):
        workbook[column + "28"] = contributor.get("contributor_name", "")
        workbook[column + "29"] = contributor.get("contributor_orcid_id", "")
        workbook[column + "30"] = contributor.get("contributor_affiliation", "")
        workbook[column + "31"] = contributor.get("contributor_role", "")
    # Return the length of the contributor array, or 1 if empty
    return max(1, len(contributor_info))


def populate_related_resource_information(workbook, soda):
    related_resource_information = soda["dataset_metadata"]["dataset_description"].get("related_resource_information", [])
    for info, column in zip(related_resource_information, excel_columns(start_index=3)):
        workbook[column + "33"] = info.get("identifier_description", "")
        workbook[column + "34"] = info.get("relation_type", "")
        workbook[column + "35"] = info.get("identifier", "")
        workbook[column + "36"] = info.get("identifier_type", "")
    # Return the length of the related resource array, or 1 if empty
    return max(1, len(related_resource_information))



def populate_participant_information(workbook, soda):
    participant_info = soda["dataset_metadata"]["dataset_description"]["participant_information"]
    workbook["D38"] = participant_info.get("number_of_subjects", 0)
    workbook["D39"] = participant_info.get("number_of_samples", 0)
    workbook["D40"] = participant_info.get("number_of_sites", 0)
    workbook["D41"] = participant_info.get("number_of_performances", 0)


def data_dictionary_information(workbook, soda):
    """
    This function is a placeholder for future implementation.
    It currently does not populate any data in the workbook.
    """
    # Placeholder for future implementation
    data_dictionary_info = soda["dataset_metadata"]["dataset_description"].get("data_dictionary_information", {})

    workbook["D43"] = data_dictionary_info.get("data_dictionary_path", "")
    workbook["D44"] = data_dictionary_info.get("data_dictionary_type", "")
    workbook["D45"] = data_dictionary_info.get("data_dictionary_description", "")

def grayout_subheaders(workbook, max_len, start_index):
    """
    Gray out sub-header rows for values exceeding 3 (SDS2.0).
    """
    headers_list = ["4", "10", "18", "23", "28"]
    columns_list = excel_columns(start_index=start_index)

    for (i, column), no in itertools.product(zip(range(2, max_len + 1), columns_list[1:]), headers_list):
        cell = workbook[column + no]
        fillColor("B2B2B2", cell)





def fillColor(color, cell):
    colorFill = PatternFill(start_color=color, end_color=color, fill_type="solid")

    cell.fill = colorFill



"""
Classes for creating manifest files for a dataset stored locally/on Pennsieve. 
"""

from os.path import join, exists, expanduser, isdir, isfile, splitext, normpath, isfile
from os import makedirs, remove, listdir
from datetime import datetime
import pathlib
import shutil 
from .manifest_import import load_metadata_to_dataframe
from ....utils import get_access_token


import pandas as pd
import requests 

from ....utils import create_request_headers, column_check, remove_high_level_folder_from_path, get_name_extension, get_dataset_id, TZLOCAL
from ....constants import PENNSIEVE_URL
userpath = expanduser("~")





def get_auto_generated_manifest_files(soda_json_structure):
    high_lvl_folders = soda_json_structure["dataset-structure"]["folders"].keys()
    manifest_folder_structure = {}
    for folder in high_lvl_folders:
        # get manifest file for each high level folder
        filePath = join(userpath, 'SODA', 'manifest_files', folder, 'manifest.xlsx')
        if(isfile(filePath)):
            manifestFilePath = normpath(filePath)
            manifest_folder_structure[folder] = manifestFilePath
    return manifest_folder_structure

def update_existing_pennsieve_manifest_files(soda_json_structure, high_level_folders, manifest_progress, manifest_path):
    """
    Updates old manifest files with new information from the dataset. Also creates new manifest files if they don't exist.
    Used in the standalone manifest workflow for Pennsieve datasets. 
    """
    dataset_id = get_dataset_id(soda_json_structure["ps-dataset-selected"]["dataset-name"])

    r = requests.get(f"{PENNSIEVE_URL}/datasets/{dataset_id}/packages", headers=create_request_headers(get_access_token()))
    r.raise_for_status()

    ds_items = r.json()["packages"]

    dataset_structure = soda_json_structure["dataset-structure"]

    # handle updating any existing manifest files on Pennsieve
    for i in ds_items:
        if i["content"]["name"] in [
            "code",
            "derivative",
            "docs",
            "primary",
            "protocol",
            "source",
        ]:
            

            limit = 100
            offset = 0
            # request the packages of that folder
            folder_name = i["content"]["name"]
            folder_collection_id = i["content"]["nodeId"]
            packageItems = []
            while True: 
                r = requests.get(f"{PENNSIEVE_URL}/packages/{folder_collection_id}?limit={limit}&offset={offset}", headers=create_request_headers(get_access_token()))
                r.raise_for_status()
                page = r.json()["children"]
                packageItems.extend(page)
                if len(page) < limit:
                    break
                offset += limit

            for j in packageItems:
                if j["content"]["name"] == "manifest.xlsx":
                    manifest_folder = join(manifest_path, folder_name)
                    if not exists(manifest_folder):
                        # create the path
                        makedirs(manifest_folder)
                    elif exists(join(manifest_folder, "manifest.xlsx")):
                        remove(join(manifest_folder, "manifest.xlsx"))

                    item_id = j["content"]["nodeId"]

                    manifest_df = load_metadata_to_dataframe(item_id, "excel", get_access_token(), column_check, 0)

                    filepath = join(
                        manifest_path, folder_name, "manifest.xlsx"
                    )

                    if folder_name not in high_level_folders:
                        continue
                        
                    high_level_folders.remove(folder_name)
                    updated_manifest_dict = update_existing_pennsieve_manifest_file(dataset_structure["folders"][folder_name], manifest_df)

                    if not exists(join(manifest_path, folder_name)):
                        # create the path
                        makedirs(join(manifest_path, folder_name))

                    new_manifest = pd.DataFrame.from_dict(updated_manifest_dict)
                    new_manifest.to_excel(filepath, index=False)

                    manifest_progress["manifest_files_uploaded"] += 1

                    # break because we only need to read the "manifest.xlsx" file in each high level folder.
                    break



def update_existing_pennsieve_manifest_file(high_level_folder, manifest_df):
    """
        Given a high level folder and the existing manifest file therein, create a new updated manifest file. 
        This manifest file needs to have files that only exist in the high level folder. Additionally, it needs to 
        retain any custom metadata that the user has added to the old manifest file for entries that still exist. 
    """

    new_manifest_dict = {'filename': [], 'timestamp': [], 'description': [], 'file type': [], 'Additional Metadata': []}
    SET_COLUMNS = ['filename', 'timestamp', 'description', 'file type', 'Additional Metadata']


    # convert the old manifest into a dictionary to optimize the lookup time
    old_manifest_dict = {x: manifest_df[x].values.tolist() for x in manifest_df}
    # old_manifest_dict = {x:manifest_df[x].values.tolist() for x in manifest_df}

    filename_idx_map = None 

    # create a mapping of filename to the idx of the row in the old_manidest_dict
    if "filename" in manifest_df:
        filename_idx_map = {x:i for i, x in enumerate(manifest_df['filename'])}
    if "File Name" in manifest_df:
        filename_idx_map = {x:i for i, x in enumerate(manifest_df['File Name'])}

    if filename_idx_map:
        for column in manifest_df.columns: 
            if column not in new_manifest_dict:
                new_manifest_dict[column] = []
                SET_COLUMNS.append(column)


    # traverse through the high level folder items
    update_existing_pennsieve_manifest_file_helper(high_level_folder, old_manifest_dict, new_manifest_dict, filename_idx_map, manifest_columns=SET_COLUMNS)

    # write the new manifest_df to the correct location as an excel file 
    return new_manifest_dict



def update_existing_pennsieve_manifest_file_helper(folder, old_manifest_dict, new_manifest_dict, filename_idx_map, manifest_columns):
    """
        Traverse through each high level folder in the dataset and update the new manifest data frame.
    """

    if "files" in folder.keys():
        for file in list(folder["files"]):
            file_path = remove_high_level_folder_from_path(folder["files"][file]["folderpath"]) + f"{file}"

            # select the row in the old manifest file that has the same file path as the file in the current folder
            # rationale: this means the file still exists in the user's dataset
            if filename_idx_map is not None:
                row_idx = filename_idx_map.get(file_path, None)
            else:
                row_idx = None

            if row_idx is None:
                for key in new_manifest_dict.keys():
                    if key == "filename":
                        new_manifest_dict["filename"].append(file_path)   
                    elif key == "timestamp":
                        new_manifest_dict["timestamp"].append(folder["files"][file]["timestamp"]), 
                    elif key == "description": 
                        new_manifest_dict["description"].append(folder["files"][file].get("description", ""))
                    elif key == "file type":
                        unused_file_name, file_extension = get_name_extension(file)
                        new_manifest_dict["file type"].append(file_extension), 
                    elif key == "Additional Metadata":
                        new_manifest_dict["Additional Metadata"].append(folder["files"][file].get("additional-metadata", ""))
                    else:
                        new_manifest_dict[key].append("")
                    

            else:
                # add the existing rows to the new manifest dictionary's arrays
                # TODO: Confirm it adds NULL/NaN if the value is empty
                for column in manifest_columns:
                    if column in old_manifest_dict:
                        new_manifest_dict[column].append(old_manifest_dict[column][row_idx])
                    else:
                        if column == "filename":
                            new_manifest_dict["filename"].append(file_path)   
                        elif column == "timestamp":
                            new_manifest_dict["timestamp"].append(folder["files"][file]["timestamp"]), 
                        elif column == "description": 
                            new_manifest_dict["description"].append(folder["files"][file].get("description", ""))
                        elif column == "file type":
                            unused_file_name, file_extension = get_name_extension(file)
                            new_manifest_dict["file type"].append(file_extension), 
                        elif column == "Additional Metadata":
                            new_manifest_dict["Additional Metadata"].append(folder["files"][file].get("additional-metadata", ""))
                        else:
                            new_manifest_dict[column].append("")

    if "folders" in folder.keys():
        for current_folder in list(folder["folders"]):
            update_existing_pennsieve_manifest_file_helper(folder["folders"][current_folder], old_manifest_dict, new_manifest_dict, filename_idx_map, manifest_columns)

    return 



def create_high_lvl_manifest_files_existing_ps_starting_point(soda_json_structure, manifest_path, manifest_progress={}):
    """
    Function to create manifest files for each high-level SPARC folder for an existing Pennsieve dataset.
    Unlike the standalone manifest generator algorithm this removes any existing manifest files and replaces them with new ones.
    Optional columns/fields a user created for the existing manifest files are not carried over into the new manifest files. 
    Args:
        soda_json_structure: soda dict with information about the dataset to be generated/modified
        high_level_folders: (optional) list of high-level folders to generate manifests for. Defaults to all primary folders.
        manifest_progress: (optional) dictionary with information about the progress of the manifest generation. Defaults to empty dictionary.
    Action:
        manifest_files_structure: dict including the local path of the manifest files
    """
    high_level_folders_present = []
    manifest_files_structure = {}
    local_timezone = TZLOCAL()


    def recursive_folder_traversal(folder, dict_folder_manifest):
        if "files" in folder.keys():
            for current_file in list(folder["files"]):
                # Construct the relative file name starting from the root of the dataset
                relative_file_name = "/".join(folder["files"][current_file]["folderpath"]) + "/" + current_file
                relative_file_name = relative_file_name.replace("\\", "/")  # Normalize to use forward slashes
                dict_folder_manifest["filename"].append(relative_file_name)

                unused_file_name, file_extension = get_name_extension(current_file)
                if file_extension == "":
                    file_extension = "None"
                dict_folder_manifest["file type"].append(file_extension)

                if "type" in folder["files"][current_file].keys():
                    if folder["files"][current_file]["type"] == "ps":
                        dict_folder_manifest["timestamp"].append(
                            folder["files"][current_file]["timestamp"]
                        )
                    elif folder["files"][current_file]["type"] == "local":
                        file_path = folder["files"][current_file]["path"]
                        filepath = pathlib.Path(file_path)
                        mtime = filepath.stat().st_mtime
                        lastmodtime = datetime.fromtimestamp(mtime).astimezone(
                            local_timezone
                        )

                        tm = lastmodtime.isoformat().replace(".", ",").replace("+00:00", "Z")
                        dict_folder_manifest["timestamp"].append(tm)
                else:
                    dict_folder_manifest["timestamp"].append("")

                if "description" in folder["files"][current_file].keys():
                    dict_folder_manifest["description"].append(
                        folder["files"][current_file]["description"]
                    )
                else:
                    dict_folder_manifest["description"].append("")

                if "additional-metadata" in folder["files"][current_file].keys():
                    dict_folder_manifest["Additional Metadata"].append(
                        folder["files"][current_file]["additional-metadata"]
                    )
                else:
                    dict_folder_manifest["Additional Metadata"].append("")

                dict_folder_manifest["entity"].append("")
                dict_folder_manifest["data modality"].append("")
                dict_folder_manifest["also in dataset"].append("")
                dict_folder_manifest["also in dataset path"].append("")
                dict_folder_manifest["data dictionary path"].append("")
                dict_folder_manifest["entity is transitive"].append("")
                

        if "folders" in folder.keys():
            for current_file in list(folder["folders"]):
                recursive_folder_traversal(
                    folder["folders"][current_file], dict_folder_manifest
                )

        return

    dataset_structure = soda_json_structure["dataset-structure"]

    # create local folder to save manifest files temporarily if the existing files are stale (i.e. not from updating existing manfiest files)
    shutil.rmtree(manifest_path) if isdir(manifest_path) else 0
    makedirs(manifest_path)

    manifestfilepath = join(manifest_path, "manifest.xlsx")

    # Initialize dict where manifest info will be stored
    dict_folder_manifest = {}
    dict_folder_manifest["filename"] = []
    dict_folder_manifest["timestamp"] = []
    dict_folder_manifest["description"] = []
    dict_folder_manifest["file type"] = []
    dict_folder_manifest["entity"] = []
    dict_folder_manifest["data modality"] = []
    dict_folder_manifest["also in dataset"] = []
    dict_folder_manifest["also in dataset path"] = []
    dict_folder_manifest["data dictionary path"] = []
    dict_folder_manifest["entity is transitive"] = []
    dict_folder_manifest["Additional Metadata"] = []

    for high_level_folder in soda_json_structure["dataset-structure"]["folders"]:
        recursive_folder_traversal(
            dataset_structure["folders"][high_level_folder], dict_folder_manifest
        )

    df = pd.DataFrame.from_dict(dict_folder_manifest)
    df.to_excel(manifestfilepath, index=None, header=True)

    # update the progress of manifest file generation
    if manifest_progress != {}:
        manifest_progress["manifest_files_uploaded"] += 1
    
    # add the path to the manifest into the structure
    manifest_files_structure[high_level_folder] = manifestfilepath

    return manifest_files_structure
    

def recursive_item_path_create(folder, path):
    if "files" in folder.keys():
        for item in list(folder["files"]):
            if "folderpath" not in folder["files"][item]:
                folder["files"][item]["folderpath"] = path[:]

    if "folders" in folder.keys():
        for item in list(folder["folders"]):
            if "folderpath" not in folder["folders"][item]:
                folder["folders"][item]["folderpath"] = path[:]
                folder["folders"][item]["folderpath"].append(item)
            recursive_item_path_create(
                folder["folders"][item], folder["folders"][item]["folderpath"][:]
            )

    return


def create_high_level_manifest_files_existing_local_starting_point(dataset_path, manifest_path):
    """
    Standalone manifest generator algorithm. Imports manifest files from a local dataset folder. 
    """
   #  soda_manifest_folder_path = join(userpath, "SODA", "manifest_files")

    if dataset_path != "":
        for high_level_fol in listdir(dataset_path):

            if high_level_fol in [
                "primary",
                "derivative",
                "docs",
                "code",
                "source",
                "protocol",
            ]:
                onlyfiles = [
                    join(dataset_path, high_level_fol, f)
                    for f in listdir(join(dataset_path, high_level_fol))
                    if isfile(join(dataset_path, high_level_fol, f))
                ]

                for file in onlyfiles:
                    p = pathlib.Path(file)
                    # create high-level folder at the temporary location
                    folderpath = join(manifest_path, high_level_fol)
                    if p.stem == "manifest":
                        if not exists(folderpath):
                            makedirs(folderpath)
                        # make copy from this manifest path to folderpath
                        shutil.copyfile(file, join(folderpath, p.name))


def create_high_level_manifest_files(soda_json_structure, manifest_path):
    """
    Function to create manifest files for each high-level SPARC folder. To be used for a local dataset.

    Args:
        soda_json_structure: soda dict with information about the dataset to be generated/modified
    Action:
        manifest_files_structure: dict including the local path of the manifest files
    """
    double_extensions = [
        ".ome.tiff",
        ".ome.tif",
        ".ome.tf2,",
        ".ome.tf8",
        ".ome.btf",
        ".ome.xml",
        ".brukertiff.gz",
        ".mefd.gz",
        ".moberg.gz",
        ".nii.gz",
        ".mgh.gz",
        ".tar.gz",
        ".bcl.gz",
    ]

    try:

        def get_name_extension(file_name):
            double_ext = False
            for ext in double_extensions:
                if file_name.find(ext) != -1:
                    double_ext = True
                    break

            ext = ""
            name = ""

            if double_ext == False:
                name = splitext(file_name)[0]
                ext = splitext(file_name)[1]
            else:
                ext = (
                    splitext(splitext(file_name)[0])[1]
                    + splitext(file_name)[1]
                )
                name = splitext(splitext(file_name)[0])[0]
            return name, ext

        def recursive_manifest_builder(
            my_folder, my_relative_path, dict_folder_manifest
        ):
            if "files" in my_folder.keys():
                for file_key, file in my_folder["files"].items():
                    dict_folder_manifest = file_manifest_entry(
                        file_key, file, my_relative_path, dict_folder_manifest
                    )

            if "folders" in my_folder.keys():
                for folder_key, folder in my_folder["folders"].items():
                    if my_relative_path:
                        relative_path = my_relative_path + "/" + folder_key
                    else:
                        relative_path = folder_key
                    dict_folder_manifest = recursive_manifest_builder(
                        folder, relative_path, dict_folder_manifest
                    )

            return dict_folder_manifest

        def file_manifest_entry(file_key, file, relative_path, dict_folder_manifest):
            # filename
            if relative_path:
                filename = relative_path + "/" + file_key
            else:
                filename = file_key
            dict_folder_manifest["filename"].append(filename)
            # timestamp
            file_type = file.get("location")
            if file_type == "local":
                file_path = file["path"]
                filepath = pathlib.Path(file_path)
                mtime = filepath.stat().st_mtime
                lastmodtime = datetime.fromtimestamp(mtime).astimezone(local_timezone)
                dict_folder_manifest["timestamp"].append(
                    lastmodtime.isoformat().replace(".", ",").replace("+00:00", "Z")
                )
            elif file_type == "bf":
                dict_folder_manifest["timestamp"].append(file["timestamp"])
            # description
            if "description" in file.keys():
                dict_folder_manifest["description"].append(file["description"])
            else:
                dict_folder_manifest["description"].append("")
            # file type
            fileextension = ""
            name_split = splitext(file_key)
            if name_split[1] == "":
                fileextension = "None"
            else:
                unused_file_name, fileextension = get_name_extension(file_key)
                # fileextension = name_split[1]
            dict_folder_manifest["file type"].append(fileextension)
            # additional metadata
            if "additional-metadata" in file.keys():
                dict_folder_manifest["Additional Metadata"].append(
                    file["additional-metadata"]
                )
            else:
                dict_folder_manifest["Additional Metadata"].append("")

            return dict_folder_manifest

        # create local folder to save manifest files temporarily (delete any existing one first)
        if "auto-generated" in soda_json_structure["manifest-files"]:
            if soda_json_structure["manifest-files"]["auto-generated"] == True:
                manifest_files_structure = (
                    get_auto_generated_manifest_files(soda_json_structure)
                )
        else:
            shutil.rmtree(manifest_path) if isdir(manifest_path) else 0
            makedirs(manifest_path)

            dataset_structure = soda_json_structure["dataset-structure"]
            local_timezone = TZLOCAL()
            manifest_files_structure = {}
            for folder_key, folder in dataset_structure["folders"].items():
                # Initialize dict where manifest info will be stored
                dict_folder_manifest = {}
                dict_folder_manifest["filename"] = []
                dict_folder_manifest["timestamp"] = []
                dict_folder_manifest["description"] = []
                dict_folder_manifest["file type"] = []
                dict_folder_manifest["Additional Metadata"] = []

                relative_path = ""
                dict_folder_manifest = recursive_manifest_builder(
                    folder, relative_path, dict_folder_manifest
                )

                # create high-level folder at the temporary location
                folderpath = join(manifest_path, folder_key)
                makedirs(folderpath)

                # save manifest file
                manifestfilepath = join(folderpath, "manifest.xlsx")
                df = pd.DataFrame.from_dict(dict_folder_manifest)
                df.to_excel(manifestfilepath, index=None, header=True)

                manifest_files_structure[folder_key] = manifestfilepath

        return manifest_files_structure

    except Exception as e:
        raise e



class ManifestWriter(object):
    """
    Writes manifest files for a dataset stored locally or on Pennsieve.
    """

    def __init__(self, soda_json_structure, path):
        """
        Constructor.
        """
        self.soda_json_structure = soda_json_structure
        # where the manifest file(s) will be written to
        self.manifest_path = path


    def write(self, soda_json_structure, ps=None):
        """
        Writes the manifest file for the dataset. Abstract.
        """
        raise NotImplementedError("Please Implement this method")


class ManifestWriterStandaloneAlgorithm(ManifestWriter):
    """
    Writes manifest files for a dataset that is stored on Pennsieve and has local changes.
    """

    def __init__(self, soda_json_structure, path):
        """
        Constructor.
        """
        super(ManifestWriterStandaloneAlgorithm, self).__init__(soda_json_structure, path)


    def write(self, soda_json_structure, ps):
        """
        Writes the manifest file for the dataset.
        """

        high_level_folders = ["code", "derivative", "docs", "primary", "protocol", "source"]

        manifest_progress = {
            "total_manifest_files": 0,
            "manifest_files_uploaded": 0,
            "finished": False
        }

        # create the manifest file
        # handle updating any existing manifest files on Pennsieve
        update_existing_pennsieve_manifest_files(soda_json_structure, high_level_folders, manifest_progress, self.manifest_path)


        recursive_item_path_create(soda_json_structure["dataset-structure"], [])

        # create manifest files from scratch for any high level folders that don't have a manifest file on Pennsieve
        create_high_lvl_manifest_files_existing_ps_starting_point(soda_json_structure, self.manifest_path, high_level_folders, manifest_progress)


class ManifestWriterStandaloneLocal(ManifestWriter):
    """
    Writes manifest files for a dataset that is stored locally.
    """

    def __init__(self, soda_json_structure, path):
        """
        Constructor.
        """
        super().__init__(soda_json_structure, path)


    def write(self, soda_json_structure, ps=None):
        """
        Writes the manifest file for the dataset.
        """

        ds_path = join(soda_json_structure["generate-dataset"]["path"], soda_json_structure["generate-dataset"]["dataset-name"])

        # TODO: Send in the correct path to the dataset using the soda_json structure
        create_high_level_manifest_files_existing_local_starting_point(ds_path, self.manifest_path)

        

class ManifestWriterNewPennsieve(ManifestWriter):
    def __init__(self, soda_json_structure, path):
        """
        Constructor.
        """
        super(ManifestWriterStandaloneAlgorithm, self).__init__(soda_json_structure, path)


    def write(self, soda_json_structure, ps):
        """
        Writes the manifest file for the dataset.
        """

        high_level_folders = ["code", "derivative", "docs", "primary", "protocol", "source"]

        manifest_progress = {
            "total_manifest_files": 0,
            "manifest_files_uploaded": 0,
            "finished": False
        }


        # create manifest files from scratch for any high level folders that don't have a manifest file on Pennsieve
        create_high_lvl_manifest_files_existing_ps_starting_point(soda_json_structure, self.manifest_path, high_level_folders, manifest_progress)



class ManifestWriterNewLocal(ManifestWriter):
    """
    Writes manifest files for a dataset that is stored locally.
    """

    def __init__(self, soda_json_structure, path):
        """
        Constructor.
        """
        super().__init__(soda_json_structure, path)


    def write(self, soda_json_structure, ps=None):
        """
        Writes the manifest file for the dataset.
        """

        create_high_level_manifest_files(soda_json_structure, self.manifest_path)


# filepath: c:\Users\aaron\pysoda\pysoda\utils\logger.py
import logging

# Create a logger for the package
logger = logging.getLogger("pysoda.utils")
logger.setLevel(logging.WARNING)  # Set the default log level


import requests
import pandas as pd
from ....constants import PENNSIEVE_URL

def load_metadata_to_dataframe(node_id, file_type, ps_or_token, usecols=None, header=0):
    """
    Given a manifests package id and its storage type - excel or csv - returns a pandas dataframe.
    IMP: Pass in the pennsieve token or pennsieve object to ps_or_token for authentication.

    Args:
        node_id (str): The id of the manifest package.
        type (str): The type of the manifest - csv or excel.
        ps_or_token (str): The pennsieve token or pennsieve object.
        usecols (list, optional): The columns to be used. Defaults to None.
        header (int, optional): The header row. Defaults to 0.
    """
    payload = {"data": {"nodeIds": [node_id]}}
    headers = { "Content-Type" : "application/json" }
    # headers = create_request_headers(ps_or_token)
    if type(ps_or_token) == str:
        r = requests.post(f"{PENNSIEVE_URL}/zipit/?api_key={ps_or_token}", json=payload, headers=headers)
    else:
        token = ps_or_token.get_user().session_token
        r = requests.post(f"{PENNSIEVE_URL}/zipit/?api_key={token}", json=payload, headers=headers)

    if file_type == "csv":
        return pd.read_csv(r.content, engine="openpyxl")
    else:
        return pd.read_excel(r.content, engine="openpyxl", usecols=usecols, header=header)


soda = {
  "starting-point": {

  },
  "ps-dataset-selected": {

  },
  "ps-accoint-selected": {

  },
  "guided-manifest-file-data": {
    "headers": [
      "filename",
      "timestamp",
      "description",
      "file type",
      "entity",
      "data modality",
      "also in dataset",
      "also in dataset path",
      "data dictionary path",
      "entity is transitive",
      "Additional Metadata"
    ],
    "data": []
  },
  "dataset-validated": False, 
  "dataset-validation-errors": {},
  "dataset-structure": {
    "folders": {
      "data": {}
    }, 
    "files": {},
    "relativePath": "/"
  },
  "dataset-metadata": {
    "description-metadata": {},
    "shared-metadata": {},
    "protocol-data": {},
    "subject-metadata": {},
    "sample-metadata": {},
    "submission-metadata": {},
    "code-metadata": {},
    "README": "",
    "CHANGES": "",
    "performances": {},
    "sites": {},
    "resources": {}
  },
  "digital-metadata": {},
  "generate-dataset": {},
  "dataset-entity-structure": {},
  "dataset-entity-obj": {},
  "standard": ""
}


from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH, SDS_FILE_SITES, SCHEMA_NAME_SITES
from .excel_utils import rename_headers, excel_columns
from openpyxl.styles import PatternFill
from os.path import join, getsize
from openpyxl import load_workbook
import shutil
from ...utils import validate_schema
from .helpers import upload_metadata_file, get_template_path

def create_excel(soda, upload_boolean, local_destination):
    source = get_template_path(SDS_FILE_SITES)

    destination = join(METADATA_UPLOAD_PS_PATH, SDS_FILE_SITES) if upload_boolean else local_destination

    shutil.copyfile(source, destination)

    wb = load_workbook(destination)
    ws1 = wb["Sheet1"]

    sites = soda["dataset_metadata"]["sites"]

    validate_schema(sites, SCHEMA_NAME_SITES)


    # get the ascii column headers
    row = 2
    ascii_headers = excel_columns(start_index=0)
    for performance in sites: 
        ws1[ascii_headers[0] + str(row)] = performance.get("site_id", "")
        ws1[ascii_headers[1] + str(row)] = performance.get("specimen_id", "")
        ws1[ascii_headers[2] + str(row)] = performance.get("site_type", "")
        ws1[ascii_headers[3] + str(row)] = performance.get("laboratory_internal_id", "")
        ws1[ascii_headers[4] + str(row)] = performance.get("coordinate_system", "")
        ws1[ascii_headers[5] + str(row)] = performance.get("coordinate_system_position", "")
        ws1[ascii_headers[6] + str(row)] = performance.get("more...", "")
        row += 1

    wb.save(destination)

    size = getsize(destination)


    ## if generating directly on Pennsieve, call upload function
    if upload_boolean:
        upload_metadata_file(SDS_FILE_SITES, soda,  destination, True)

    return {"size": size}
    






from string import ascii_uppercase
import itertools
from openpyxl.styles import PatternFill, Font



def rename_headers(workbook, max_len, start_index):
  """
  Rename header columns if values exceed 3. Change Additional Values to Value 4, 5,...
  Adds styling to the column headers as well.
  """

  columns_list = excel_columns(start_index=start_index)
  if max_len >= start_index:
      workbook[columns_list[0] + "1"] = "Value"
      for i, column in zip(range(2, max_len + 1), columns_list[1:]):

          workbook[column + "1"] = f"Value {str(i)}"
          cell = workbook[column + "1"]

          blueFill = PatternFill(
              start_color="9CC2E5", end_color="9CC2E5", fill_type="solid"
          )

          font = Font(bold=True)
          cell.fill = blueFill
          cell.font = font

  else:
      delete_range = len(columns_list) - max_len
      workbook.delete_cols(4 + max_len, delete_range)



def excel_columns(start_index=0):
    """
    NOTE: does not support more than 699 contributors/links
    """
    single_letter = list(ascii_uppercase[start_index:])
    two_letter = [a + b for a, b in itertools.product(ascii_uppercase, ascii_uppercase)]
    return single_letter + two_letter



from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH, SDS_FILE_MANIFEST, SCHEMA_NAME_MANIFEST
from .excel_utils import rename_headers, excel_columns
from openpyxl.styles import Font, PatternFill
from os.path import join, getsize
from openpyxl import load_workbook
import shutil
from ...utils import validate_schema, get_schema_path
from .helpers import upload_metadata_file
from .helpers import get_template_path
import os


from json import load as json_load



def create_excel(soda, upload_boolean, local_destination):
    source = get_template_path("manifest.xlsx")
    destination = join(METADATA_UPLOAD_PS_PATH, SDS_FILE_MANIFEST) if upload_boolean else local_destination
    shutil.copyfile(source, destination)
    wb = load_workbook(destination)
    ws1 = wb["Sheet1"]
    manifest = soda["dataset_metadata"]["manifest_file"]
    # validate_schema(manifest, SCHEMA_NAME_MANIFEST)
    ascii_headers = excel_columns(start_index=0)
    custom_headers_to_column = {}

    orangeFill = PatternFill(
        start_color="FFD965", end_color="FFD965", fill_type="solid"
    )

    # Load schema to get standard headers
    schema_path =get_schema_path(SCHEMA_NAME_MANIFEST)
    with open(schema_path, "r") as f:
        schema = json_load(f)
    # The schema is an array, so get the first item's properties
    item_schema = schema["items"][0]
    standard_headers = list(item_schema["properties"].keys())

    # Write standard headers to the first row
    for idx, header in enumerate(standard_headers):
        ws1[ascii_headers[idx] + "1"] = header.replace("_", " ")
        ws1[ascii_headers[idx] + "1"].font = Font(bold=True, size=12, name="Calibri")

    row = 2
    for entry in manifest:
        # Write standard fields
        for idx, header in enumerate(standard_headers):
            value = entry.get(header, "")
            if isinstance(value, list):
                value = " ".join(value)
            ws1[ascii_headers[idx] + str(row)] = value
            ws1[ascii_headers[idx] + str(row)].font = Font(bold=False, size=11, name="Arial")

        # Handle custom fields
        for field_name, field_value in entry.items():
            if field_name in standard_headers:
                continue
            if field_name not in custom_headers_to_column:
                custom_headers_to_column[field_name] = len(custom_headers_to_column) + len(standard_headers)
                col_idx = custom_headers_to_column[field_name]
                ws1[ascii_headers[col_idx] + "1"] = field_name
                ws1[ascii_headers[col_idx] + "1"].fill = orangeFill
                ws1[ascii_headers[col_idx] + "1"].font = Font(bold=True, size=12, name="Calibri")
            col_idx = custom_headers_to_column[field_name]
            ws1[ascii_headers[col_idx] + str(row)] = field_value
            ws1[ascii_headers[col_idx] + str(row)].font = Font(bold=False, size=11, name="Arial")
        row += 1

    # Rename additional metadata header to Additional Metadata header
    # ws1[ascii_headers[len(standard_headers)] + "1"] = "Additional Metadata"

    wb.save(destination)
    size = getsize(destination)
    if upload_boolean:
        upload_metadata_file(SDS_FILE_MANIFEST, soda, destination, True)

    return {"size": size}




def load_existing_manifest_file(manifest_file_path):
   # check that a file exists at the path 
    if not os.path.exists(manifest_file_path):
        raise FileNotFoundError(f"Manifest file not found at {manifest_file_path}")
    
   # load the xlsx file and store its first row as a headers array and the rest of the rows in a data key 
    wb = load_workbook(manifest_file_path)
    ws1 = wb["Sheet1"]
    headers = []
    data = []

    for row in ws1.iter_rows(min_row=1, max_row=1, values_only=True):
        headers = list(row)

    for row in ws1.iter_rows(min_row=2, values_only=True):
        # grab the item in row 5
        new_row = []
        for col_data in row:
            if isinstance(col_data, list):
                # space separate the list into a string
                col_data = " ".join(col_data)
            new_row.append(col_data)
        data.append(new_row)

    return {"headers": headers, "data": data}








import requests
from .. import logger
from ...constants import PENNSIEVE_2_URL
from ...utils import get_access_token, create_request_headers




def get_files_for_manifest(manifest_id, limit, continuation_token=None):
    """
    Get the number of verified files in an upload manifest. For a file to be verified its status must be 
    one of the following: VERIFIED | FAILED | FINALIZED.
    """

    global logger


    if continuation_token is None or continuation_token == "":
        r = requests.get(f"{PENNSIEVE_2_URL}/manifest/files?manifest_id={manifest_id}&limit={limit}", headers=create_request_headers(get_access_token()))
    else:
        r = requests.get(f"{PENNSIEVE_2_URL}/manifest/files?manifest_id={manifest_id}&limit={limit}&continuation_token={continuation_token}", headers=create_request_headers(get_access_token()))
    r.raise_for_status()
    return r.json()



def get_upload_manifests(dataset_id):
    """
    Get the ids of all upload manifests that have been initiated by the Pennsieve Agent.
    """

    global logger

    r = requests.get(f"{PENNSIEVE_2_URL}/manifest?dataset_id={dataset_id}", headers=create_request_headers(get_access_token()))
    r.raise_for_status()

    return r.json()


import os.path
import requests
from .authentication import create_request_headers
from ..constants import PENNSIEVE_URL


# check for non-empty fields (cells)
def column_check(x):
    return "unnamed" not in x.lower()


# obtain Pennsieve S3 URL for an existing metadata file
def returnFileURL(ps, item_id):
    r = requests.get(f"{PENNSIEVE_URL}/packages/{item_id}/view", headers=create_request_headers(ps))
    r.raise_for_status()

    file_details = r.json()
    file_id = file_details[0]["content"]["id"]
    r = requests.get(
        f"{PENNSIEVE_URL}/packages/{item_id}/files/{file_id}", headers=create_request_headers(ps)
    )
    r.raise_for_status()

    file_url_info = r.json()
    return file_url_info["url"]


def remove_high_level_folder_from_path(paths):
    """
        Remove the high level folder from the path. This is necessary because the high level folder is not included in the manifest file name entry.
    """

    return "" if len(paths) == 1 else "/".join(paths[1:]) + "/"




double_extensions = [
    ".ome.tiff",
    ".ome.tif",
    ".ome.tf2,",
    ".ome.tf8",
    ".ome.btf",
    ".ome.xml",
    ".brukertiff.gz",
    ".mefd.gz",
    ".moberg.gz",
    ".nii.gz",
    ".mgh.gz",
    ".tar.gz",
    ".bcl.gz",
]


def get_name_extension(file_name):
    double_ext = False
    for ext in double_extensions:
        if file_name.find(ext) != -1:
            double_ext = True
            break

    ext = ""
    name = ""

    if double_ext == False:
        name = os.path.splitext(file_name)[0]
        ext = os.path.splitext(file_name)[1]
    else:
        ext = (
            os.path.splitext(os.path.splitext(file_name)[0])[1]
            + os.path.splitext(file_name)[1]
        )
        name = os.path.splitext(os.path.splitext(file_name)[0])[0]
    return name, ext


from .constants import METADATA_UPLOAD_PS_PATH, TEMPLATE_PATH, SDS_FILE_CODE_DESCRIPTION, SCHEMA_NAME_CODE_DESCRIPTION
from .excel_utils import rename_headers, excel_columns
from openpyxl.styles import PatternFill
from os.path import join, getsize
from openpyxl import load_workbook
import shutil
from ...utils import validate_schema
from .helpers import upload_metadata_file, get_template_path



# TODO: Handle optional entries when coupled with provided entries
# TODO: Handle extending columns and filling with color when more entries are provided than the template default handles
def create_excel(soda, upload, local_destination):
    source = get_template_path(SDS_FILE_CODE_DESCRIPTION)
    destination = join(METADATA_UPLOAD_PS_PATH, SDS_FILE_CODE_DESCRIPTION) if upload else local_destination
    shutil.copyfile(source, destination)

    validate_schema(soda["dataset_metadata"]["code_description"], SCHEMA_NAME_CODE_DESCRIPTION)

    wb = load_workbook("./" +destination)
    print(wb.sheetnames)
    ws1 = wb[wb.sheetnames[0]]

    populate_input_output_information(ws1, soda)

    populate_basic_information(ws1, soda)

    populate_ten_simple_rules(ws1, soda)

    wb.save(destination)

    size = getsize(destination)

    ## if generating directly on Pennsieve, then call upload function and then delete the destination path
    if upload:
        upload_metadata_file(
            SDS_FILE_CODE_DESCRIPTION, soda, destination, True
        )

    return {"size": size}





# TODO: Handle optional entries
def populate_input_output_information(ws1, soda):
    # populate from row 27 and column 4 up to column n, depending upon the amount of items in the array for each input output information entry
    input_output_information = soda["dataset_metadata"]["input_output_information"]

    row = 27

    excel_ascii = excel_columns(start_index=3)[0]
    ws1[excel_ascii + str(row)] = input_output_information["number_of_inputs"]

    for input, column in zip(input_output_information["inputs"], excel_columns(start_index=3)):
        row = 28
        ws1[column + str(row)] = input["input_parameter_name"]
        ws1[column + str(row + 1)] = input["input parameter type"]
        ws1[column + str(row + 2)] = input["input_parameter_description"]
        ws1[column + str(row + 3)] = input["input_units"]
        ws1[column + str(row + 4)] = input["input_default_value"]

    # populate number of outputs into row 34
    row = 34
    ws1[excel_ascii + str(row)] = input_output_information["number_of_outputs"]

    # populate the outputs from row 35 - 39
    for output, column in zip(input_output_information["outputs"], excel_columns(start_index=3)):
        row = 35
        ws1[column + str(row)] = output["output_parameter_name"]
        ws1[column + str(row + 1)] = output["output_parameter_type"]
        ws1[column + str(row + 2)] = output["output_parameter_description"]
        ws1[column + str(row + 3)] = output["output_units"]
        ws1[column + str(row + 4)] = output["output_default_value"]


def populate_basic_information(ws1, soda):
    basic_information = soda["dataset_metadata"]["basic_information"]

    # fill out basic information from row 2 - 5 starting from col 3
    row = 2
    for info, column in zip(basic_information, excel_columns(start_index=3)):
        ws1[column + str(row)] = info["RRID_term"]
        ws1[column + str(row + 1)] = info["RRID_identifier"]
        ws1[column + str(row + 2)] = info["ontology_term"]
        ws1[column + str(row + 3)] = info["ontology_identifier"]


def populate_ten_simple_rules(ws1, soda):
    ten_simple_rules = soda["dataset_metadata"]["ten_simple_rules"]
    row = 8
    ascii_cols = excel_columns(start_index=3)
    for _, rule in ten_simple_rules.items():
        ws1[ascii_cols[0] + str(row)] = rule.get("Link", "")
        ws1[ascii_cols[1] + str(row)] = rule.get("Rating", "")
        ws1[ascii_cols[2] + str(row)] = rule.get("Target", "")
        ws1[ascii_cols[3] + str(row)] = rule.get("Target Justification", "")
        ws1[ascii_cols[4] + str(row)] = rule.get("Text", "")
        row += 1






    

